/*
 * SPDX-FileCopyrightText: 2024-2026 Sébastien Helleu <flashcode@flashtux.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of WeeChat, the extensible chat client.
 *
 * WeeChat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * WeeChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
 */

/* Test hook info functions */

#include "CppUTest/TestHarness.h"

#include "tests.h"

extern "C"
{
#include "src/core/weechat.h"
}

TEST_GROUP(HookInfo)
{
};

/*
 * Test functions:
 *   hook_info_get_description
 */

TEST(HookInfo, GetDescription)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_info
 */

TEST(HookInfo, Info)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_info_get
 */

TEST(HookInfo, Get)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_info_free_data
 */

TEST(HookInfo, FreeData)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_info_add_to_infolist
 */

TEST(HookInfo, AddToInfolist)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_info_print_log
 */

TEST(HookInfo, PrintLog)
{
    /* TODO: write tests */
}
