/*
 * SPDX-FileCopyrightText: 2024-2026 Sébastien Helleu <flashcode@flashtux.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of WeeChat, the extensible chat client.
 *
 * WeeChat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * WeeChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
 */

/* Test hook url functions */

#include "CppUTest/TestHarness.h"

#include "tests.h"

extern "C"
{
#include "src/core/weechat.h"
}

TEST_GROUP(HookUrl)
{
};

/*
 * Test functions:
 *   hook_url_get_description
 */

TEST(HookUrl, GetDescription)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_hashtable_map_cb
 */

TEST(HookUrl, HashtableMapCb)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_run_callback
 */

TEST(HookUrl, RunCallback)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_thread_cleanup
 */

TEST(HookUrl, ThreadCleanup)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_transfer_thread
 */

TEST(HookUrl, TransferThread)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_timer_cb
 */

TEST(HookUrl, TimerCb)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_transfer
 */

TEST(HookUrl, Transfer)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url
 */

TEST(HookUrl, Url)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_free_data
 */

TEST(HookUrl, FreeData)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_add_to_infolist
 */

TEST(HookUrl, AddToInfolist)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_url_print_log
 */

TEST(HookUrl, PrintLog)
{
    /* TODO: write tests */
}
