/*
 * SPDX-FileCopyrightText: 2018-2026 Sébastien Helleu <flashcode@flashtux.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of WeeChat, the extensible chat client.
 *
 * WeeChat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * WeeChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
 */

/* Test hook functions */

#include "CppUTest/TestHarness.h"

extern "C"
{
}

TEST_GROUP(CoreHook)
{
};

/*
 * Test functions:
 *   hook_init
 */

TEST(CoreHook, Init)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_search_type
 */

TEST(CoreHook, SearchType)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_find_pos
 */

TEST(CoreHook, FindPos)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_add_to_list
 */

TEST(CoreHook, AddToList)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_remove_from_list
 */

TEST(CoreHook, RemoveFromList)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_remove_deleted
 */

TEST(CoreHook, RemoveDeleted)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_init_data
 */

TEST(CoreHook, InitData)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_valid
 */

TEST(CoreHook, Valid)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_exec_start
 */

TEST(CoreHook, ExecStart)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_exec_end
 */

TEST(CoreHook, ExecEnd)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_callback_start
 */

TEST(CoreHook, CallbackStart)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_callback_end
 */

TEST(CoreHook, CallbackEnd)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_get_description
 */

TEST(CoreHook, GetDescription)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_set
 */

TEST(CoreHook, Set)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_timer_clean_process_cb
 */

TEST(CoreHook, TimerCleanProcessCb)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_schedule_clean_process
 */

TEST(CoreHook, ScheduleCleanProcess)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   unhook
 */

TEST(CoreHook, Unhook)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   unhook_all_plugin
 */

TEST(CoreHook, UnhookAllPlugin)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   unhook_all
 */

TEST(CoreHook, UnhookAll)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_add_to_infolist_pointer
 */

TEST(CoreHook, AddToInfolistPointer)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_add_to_infolist_type
 */

TEST(CoreHook, AddToInfolistType)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_add_to_infolist
 */

TEST(CoreHook, AddToInfolist)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_print_log
 */

TEST(CoreHook, PrintLog)
{
    /* TODO: write tests */
}
